import { IterResult } from '../utils'
import { Deserialize, IDeserializer, IterableAccess, IVisitor, MapAccess } from './interface'

export class ForwardMapAccess extends MapAccess {
  private readonly keys: string[]
  private readonly values: any[]

  private kindex: number = 0
  private vindex: number = 0

  constructor(keys: string[], values: any[]) {
    super()
    this.keys = keys
    this.values = values
  }

  static fromObject(obj: object): ForwardMapAccess {
    return new ForwardMapAccess(
      Object.keys(obj),
      Object.values(obj)
    )
  }

  nextKeySeed<T, K extends Deserialize>(_seed: K): IteratorResult<T> {
    return this.nextKey()
  }

  nextValueSeed<T, V extends Deserialize>(_seed: V): IteratorResult<T> {
    return this.nextValue()
  }

  nextKey<T>(): IteratorResult<T> {
    if (this.kindex < this.keys.length) {
      return IterResult.Next(this.keys[this.kindex++]) as IteratorResult<T>
    } else {
      return IterResult.Done()
    }
  }

  nextValue<V>(): IteratorResult<V> {
    if (this.vindex < this.values.length) {
      return IterResult.Next(this.values[this.vindex++]) as IteratorResult<V>
    } else {
      return IterResult.Done()
    }
  }
}

export class ForwardIterableAccess extends IterableAccess {
  private readonly items: any[]
  private index: number = 0

  constructor(items: any[]) {
    super()
    this.items = items
  }

  nextElement<T>(): IteratorResult<T> {
    if (this.index < this.items.length) {
      return IterResult.Next(this.items[this.index++]) as IteratorResult<T>
    } else {
      return IterResult.Done()
    }
  }
}

export class Forward implements IDeserializer {
  private readonly value: any

  constructor(value: any) {
    this.value = value
  }

  static with(value: any): Forward {
    return new this(value)
  }

  deserializeAny<T, V extends IVisitor<T>>(_visitor: V): T {
    throw new Error("Can't forward to deserializeAny")
  }

  deserializeBoolean<T, V extends IVisitor<T>>(visitor: V): T {
    return visitor.visitBoolean(this.value)
  }

  deserializeNumber<T, V extends IVisitor<T>>(visitor: V): T {
    return visitor.visitNumber(this.value)
  }

  deserializeBigInt<T, V extends IVisitor<T>>(visitor: V): T {
    return visitor.visitBigInt(this.value)
  }

  deserializeString<T, V extends IVisitor<T>>(visitor: V): T {
    return visitor.visitString(this.value)
  }

  deserializeSymbol<T, V extends IVisitor<T>>(visitor: V): T {
    return visitor.visitSymbol(this.value)
  }

  deserializeNull<T, V extends IVisitor<T>>(visitor: V): T {
    return visitor.visitNull()
  }

  deserializeObject<T, V extends IVisitor<T>>(visitor: V): T {
    return visitor.visitObject(ForwardMapAccess.fromObject(this.value))
  }

  deserializeIterable<T, V extends IVisitor<T>>(visitor: V): T {
    return visitor.visitIterable(new ForwardIterableAccess(this.value))
  }

  deserializeFunction<T, V extends IVisitor<T>>(_visitor: V): T {
    throw new Error('Method not implemented.')
  }
}

