import { CaseConvention, convertCase } from '../case'
import { Deserialize } from '../de'
import { isFunction, isNumber, isString, Nullable } from '../utils'
import { Serialize } from './interface'

export interface RenameOptions {
  serialize?: string
  deserialize?: string
}

export interface RenameAllOptions {
  serialize?: CaseConvention
  deserialize?: CaseConvention
}

export interface ContainerOptions {
  rename?: RenameOptions | string
  renameAll?: RenameAllOptions | CaseConvention
  default?: () => any
  denyUnknownFields?: boolean
  tag?: string
  untagged?: boolean
}

interface Predicate {
  <T>(value: T): boolean
}

export interface SkipOptions {
  serialize?: Predicate | boolean
  deserialize?: Predicate | boolean
}
export interface PropertyOptions {
  alias?: string | string[]
  default?: () => any
  flatten?: boolean
  rename?: RenameOptions | string
  skip?: SkipOptions | Predicate | boolean
  serializeWith?: Serialize<any>
  deserializeWith?: Deserialize
}

export const Stage = Object.freeze({
  Serialize: 0,
  Deserialize: 1
} as const)

export type Stage = typeof Stage[keyof typeof Stage]

export class SerdeOptions {
  readonly target: Nullable<any>
  readonly container: ContainerOptions
  readonly properties: Map<string, PropertyOptions>

  constructor(target: any, container: ContainerOptions = {}, properties: Map<string, PropertyOptions> = new Map()) {
    this.target = target
    this.container = container
    this.properties = properties
  }

  private getClassName(stage: Stage, defaultName: string = this.target?.constructor?.name) {
    if (isString(this.container.rename)) {
      return this.container.rename
    } else if (stage === Stage.Serialize && isString(this.container.rename?.serialize)) {
      return this.container.rename.serialize
    } else if (stage === Stage.Deserialize && isString(this.container.rename?.deserialize)) {
      return this.container.rename.serialize
    } else {
      return defaultName
    }
  }

  getSerializedClassName(defaultName: string = this.target?.constructor?.name) {
    return this.getClassName(Stage.Serialize, defaultName)
  }

  getDeserializedClassName(defaultName: string = this.target?.constructor?.name) {
    return this.getClassName(Stage.Deserialize, defaultName)
  }

  private applyPropertyCase(stage: Stage, property: string) {
    if (isNumber(this.container.renameAll)) {
      return convertCase(property, this.container.renameAll)
    } else if (stage === Stage.Serialize && isNumber(this.container.renameAll?.serialize)) {
      return convertCase(property, this.container.renameAll.serialize)
    } else if (stage === Stage.Deserialize && isNumber(this.container.renameAll?.deserialize)) {
      return convertCase(property, this.container.renameAll.deserialize)
    } else {
      return property
    }
  }

  private getPropertyName(stage: Stage, property: string) {
    const options = this.properties.get(property)

    if (options == null) {
      return property
    } else if (isString(options.rename)) {
      return options.rename
    } else if (stage == Stage.Serialize && isString(options.rename?.serialize)) {
      return options.rename.serialize
    } else if (stage == Stage.Deserialize && isString(options.rename?.deserialize)) {
      return options.rename.deserialize
    } else {
      return property
    }
  }

  getSerializationPropertyName(property: string) {
    const name = this.getPropertyName(Stage.Serialize, property)
    return this.applyPropertyCase(Stage.Serialize, name)
  }

  getDeserializationPropertyName(property: string) {
    const name = this.getPropertyName(Stage.Deserialize, property)
    return this.applyPropertyCase(Stage.Deserialize, name)
  }

  private shouldSkip(stage: Stage, property: string, value: any) {
    const options = this.properties.get(property)
    if (options == null) {
      return false
    } else if (typeof options.skip === 'boolean') {
      return options.skip
    } else if (isFunction(options.skip)) {
      return options.skip(value)
    } else if (stage === Stage.Serialize && typeof options.skip?.serialize === 'boolean') {
      return options.skip.serialize
    } else if (stage === Stage.Serialize && isFunction(options.skip?.serialize)) {
      return options.skip.serialize(value)
    } else if (stage === Stage.Deserialize && typeof options.skip?.deserialize === 'boolean') {
      return options.skip.deserialize
    } else if (stage === Stage.Deserialize && isFunction(options.skip?.deserialize)) {
      return options.skip.deserialize(value)
    } else {
      return false
    }
  }

  shouldSerialize(property: string, value: any) {
    return !this.shouldSkip(Stage.Serialize, property, value)
  }

  shouldDeserialize(property: string, value: any) {
    return !this.shouldSkip(Stage.Deserialize, property, value)
  }

  defaultFor(property: string) {
    const options = this.properties.get(property)
    if (options != null && isFunction(options.default)) {
      return options.default()
    } else if (isFunction(this.container.default)) {
      return this.container.default()
    }
  }
}

