import { SerdeOptions } from './decorator'
import { ISerializeObject, Serializer } from './interface'
import { isPlainObject, Nullable } from '../utils'

class UnhandledTypeError extends TypeError {
  constructor(serializer: Serializer<unknown>, value: any) {
    super(`unhandled type: '${typeof value}' for serializer ${serializer.constructor.name}`)
  }
}

function serializeObject<T, V extends object, S extends ISerializeObject<T>>(serializer: S, obj: V, options?: SerdeOptions): T {
  for (const key in obj) {
    serializer.serializeEntry(key, obj[key])
  }

  return serializer.end()
}

function serializeClass<T, V extends object, S extends Serializer<T>>(serializer: S, value: V, options?: SerdeOptions): T {
  const name = value.constructor.name
  const ser = serializer.serializeClass(name)
  return serializeObject(ser, value, options)
}

const defaultGetter = (value: any): Nullable<SerdeOptions> => {
  return value[Symbol.metadata]?.serde
}

export function serialize<T, V, S extends Serializer<T>>(serializer: S, value: V, optionsGetter: (value: V) => Nullable<SerdeOptions> = defaultGetter): T {
  switch (typeof value) {
    case 'string': return serializer.serializeString(value)
    case 'number': return serializer.serializeNumber(value)
    case 'bigint': return serializer.serializeBigInt(value)
    case 'boolean': return serializer.serializeBoolean(value)
    case 'symbol': return serializer.serializeSymbol(value)
    case 'undefined': return serializer.serializeNull()
    case 'object':
      const options = optionsGetter(value)
      switch (true) {
        case value == null: return serializer.serializeNull()
        case !isPlainObject(value): return serializeClass(serializer, value, options)
        default: return serializeObject(serializer.serializeObject(), value, options)
      }
    default: throw new UnhandledTypeError(serializer, value)
  }
}

